;; Generated package description from kind-icon.el  -*- no-byte-compile: t -*-
(define-package "kind-icon" "0.1.7.0.20221107.212930" "Completion kind icons" '((emacs "27.1") (svg-lib "0")) :commit "02b24406d41739020638205c2a06710fe1d48381" :authors '(("J.D. Smith" . "jdtsmith@gmail.com")) :maintainer '("J.D. Smith" . "jdtsmith@gmail.com") :keywords '("completion") :url "https://github.com/jdtsmith/kind-icon")
